@@ -12,7 +12,7 @@
 #include <crypto/scatterwalk.h>
 #include <asm/unaligned.h>
 
-#ifdef CONFIG_X86_64
+#if defined(CONFIG_X86_64)
 #include <asm/cpufeature.h>
 #include <asm/processor.h>
 #ifdef CONFIG_AS_SSSE3
@@ -37,8 +37,22 @@ void chacha20poly1305_fpu_init(void)
 	chacha20poly1305_use_ssse3 = boot_cpu_has(X86_FEATURE_SSSE3);
 	chacha20poly1305_use_avx2 = boot_cpu_has(X86_FEATURE_AVX) && boot_cpu_has(X86_FEATURE_AVX2);
 }
+#elif IS_ENABLED(CONFIG_KERNEL_MODE_NEON)
+#include <asm/hwcap.h>
+#include <asm/neon.h>
+asmlinkage void chacha20_asm_block_xor_neon(u32 *state, u8 *dst, const u8 *src);
+asmlinkage void chacha20_asm_4block_xor_neon(u32 *state, u8 *dst, const u8 *src);
+static bool chacha20poly1305_use_neon __read_mostly = false;
+void __init chacha20poly1305_fpu_init(void)
+{
+#if defined(CONFIG_ARM64)
+	chacha20poly1305_use_neon = elf_hwcap & HWCAP_ASIMD;
+#elif defined(CONFIG_ARM)
+	chacha20poly1305_use_neon = elf_hwcap & HWCAP_NEON;
+#endif
+}
 #else
-void chacha20poly1305_fpu_init(void) { }
+void __init chacha20poly1305_fpu_init(void) { }
 #endif
 
 #define CHACHA20_IV_SIZE	16
@@ -257,13 +271,16 @@ static void chacha20_crypt(struct chacha20_ctx *ctx, u8 *dst, const u8 *src, uns
 	u8 buf[CHACHA20_BLOCK_SIZE];
 
 	if (!have_simd
-#ifdef CONFIG_X86_64
+#if defined(CONFIG_X86_64)
 		|| !chacha20poly1305_use_ssse3
+
+#elif IS_ENABLED(CONFIG_KERNEL_MODE_NEON)
+		|| !chacha20poly1305_use_neon
 #endif
 	)
 		goto no_simd;
 
-#ifdef CONFIG_X86_64
+#if defined(CONFIG_X86_64)
 #ifdef CONFIG_AS_AVX2
 	if (chacha20poly1305_use_avx2) {
 		while (bytes >= CHACHA20_BLOCK_SIZE * 8) {
@@ -297,6 +314,27 @@ static void chacha20_crypt(struct chacha20_ctx *ctx, u8 *dst, const u8 *src, uns
 	}
 	return;
 #endif
+#elif IS_ENABLED(CONFIG_KERNEL_MODE_NEON)
+	while (bytes >= CHACHA20_BLOCK_SIZE * 4) {
+		chacha20_asm_4block_xor_neon(ctx->state, dst, src);
+		bytes -= CHACHA20_BLOCK_SIZE * 4;
+		src += CHACHA20_BLOCK_SIZE * 4;
+		dst += CHACHA20_BLOCK_SIZE * 4;
+		ctx->state[12] += 4;
+	}
+	while (bytes >= CHACHA20_BLOCK_SIZE) {
+		chacha20_asm_block_xor_neon(ctx->state, dst, src);
+		bytes -= CHACHA20_BLOCK_SIZE;
+		src += CHACHA20_BLOCK_SIZE;
+		dst += CHACHA20_BLOCK_SIZE;
+		ctx->state[12]++;
+	}
+	if (bytes) {
+		memcpy(buf, src, bytes);
+		chacha20_asm_block_xor_neon(ctx->state, buf, buf);
+		memcpy(dst, buf, bytes);
+	}
+	return;
 #endif
 
 no_simd:
@@ -421,7 +459,7 @@ static void poly1305_simd_mult(u32 *a, const u32 *b)
 	memset(m, 0, sizeof(m));
 	/* The poly1305 block function adds a hi-bit to the accumulator which
 	 * we don't need for key multiplication; compensate for it. */
-	a[4] -= 1 << 24;
+	a[4] -= 1U << 24;
 	poly1305_asm_block_sse2(a, m, b, 1);
 }
 
@@ -485,7 +523,7 @@ static void poly1305_update(struct poly1305_ctx *ctx, const u8 *src, unsigned in
 				poly1305_simd_blocks(ctx, ctx->buf, POLY1305_BLOCK_SIZE);
 			else
 #endif
-				poly1305_generic_blocks(ctx, ctx->buf, POLY1305_BLOCK_SIZE, 1 << 24);
+				poly1305_generic_blocks(ctx, ctx->buf, POLY1305_BLOCK_SIZE, 1U << 24);
 			ctx->buflen = 0;
 		}
 	}
@@ -496,7 +534,7 @@ static void poly1305_update(struct poly1305_ctx *ctx, const u8 *src, unsigned in
 			bytes = poly1305_simd_blocks(ctx, src, srclen);
 		else
 #endif
-			bytes = poly1305_generic_blocks(ctx, src, srclen, 1 << 24);
+			bytes = poly1305_generic_blocks(ctx, src, srclen, 1U << 24);
 		src += srclen - bytes;
 		srclen = bytes;
 	}
@@ -536,10 +574,10 @@ static void poly1305_finish(struct poly1305_ctx *ctx, u8 *dst)
 
 	/* compute h + -p */
 	g0 = h0 + 5;
-	g1 = h1 + (g0 >> 26);             g0 &= 0x3ffffff;
-	g2 = h2 + (g1 >> 26);             g1 &= 0x3ffffff;
-	g3 = h3 + (g2 >> 26);             g2 &= 0x3ffffff;
-	g4 = h4 + (g3 >> 26) - (1 << 26); g3 &= 0x3ffffff;
+	g1 = h1 + (g0 >> 26);              g0 &= 0x3ffffff;
+	g2 = h2 + (g1 >> 26);              g1 &= 0x3ffffff;
+	g3 = h3 + (g2 >> 26);              g2 &= 0x3ffffff;
+	g4 = h4 + (g3 >> 26) - (1U << 26); g3 &= 0x3ffffff;
 
 	/* select h if h < p, or h + -p if h >= p */
 	mask = (g4 >> ((sizeof(u32) * 8) - 1)) - 1;
